;EasyCodeName=CreateProcessAndRedirectOutput,1
DoCallback Macro callback,event,pparam,param
	Cmp callback, NULL
	Je @F
	Push param
	Push pparam
	Push event
	Call callback
@@:
	ExitM
EndM

;******************************************************************************
; CreateProcessAndRedirectOutput: Runs a process and redirect it's output
;
;	lpszScript: A commandline for a bat, vbs, js, exe or other file
;
;	lpCallback: Function pointer
;
;	Callback PROTO dwEvent:DWORD,lpParam:DWORD,dwParam:DWORD
;
;	If dwEvent == 1 Then
;		description: callback when reading from console output
;		lpParam: a buffer containing the console output
;		dwParam: the size of lpParam
;	ElseIf dwEvent == 2
;		description: the process return value
;		lpParam: the command line
;		dwParam: the return value of the process
;	ElseIf dwEvent == 3
;		description: error
;		lpParam: NULL
;		dwParam: 0
;	EndIf
.Code

CreateProcessAndRedirectOutput Proc Public lpszScript:DWord, lpCallback:DWord
	Local dwRet:DWord
	Local dwRead:DWord
	Local lpBuffer:DWord
	Local hCurrentProcess:HANDLE
	Local hChildProcess:HANDLE
	Local hOutReadTmp:HANDLE
	Local hOutRead:HANDLE
	Local hOutWrite:HANDLE
	Local hErrWrite:HANDLE
	Local sa:SECURITY_ATTRIBUTES

	Mov dwRet, 0

	; Initialize for cleanup,...
	Mov lpBuffer, NULL
	Mov hChildProcess, NULL
	Mov hOutReadTmp, NULL
	Mov hOutRead, NULL
	Mov hOutWrite, NULL
	Mov hErrWrite, NULL

	; Get the current process handle
	Invoke GetCurrentProcess
	Mov hCurrentProcess, Eax

	; Set up the security attributes Struct.
	Mov sa.nLength, SizeOf SECURITY_ATTRIBUTES
	Mov sa.lpSecurityDescriptor, NULL
	Mov sa.bInheritHandle, TRUE

	; Create the child output pipe.
	Invoke CreatePipe, Addr hOutReadTmp, Addr hOutWrite, Addr sa, 0
	Or Eax, Eax
	Jz proc_error

	; Create a new output read handle. Set the Properties to FALSE.
	; Otherwise, the child inherits the properties and, as a result,
	; non-closeable handles to the pipes are created.
	Invoke DuplicateHandle, hCurrentProcess, hOutReadTmp, hCurrentProcess, Addr hOutRead, 0, FALSE, DUPLICATE_SAME_ACCESS
	Or Eax, Eax
	Jz proc_error

	; Create a DUPLICATE of the output write handle For the Std error
	; write handle. This is necessary in case the child application
	; closes one of its std output handles.
	Invoke DuplicateHandle, hCurrentProcess, hOutWrite, hCurrentProcess, Addr hErrWrite, 0, TRUE, DUPLICATE_SAME_ACCESS
	Or Eax, Eax
	Jz proc_error

	; Close inheritable copy of the handle you do Not want to be inherited.
	Invoke CloseHandle, hOutReadTmp
	Mov hOutReadTmp, NULL
	Or Eax, Eax
	Jz proc_error

	Invoke LaunchRedirectedProcess, lpszScript, hOutWrite, NULL, hErrWrite, Addr hChildProcess
	Or Eax, Eax
	Jz proc_error
	
	Invoke CloseHandle, hOutWrite
	Mov hOutWrite, NULL
	Or Eax, Eax
	Jz proc_error
	Invoke CloseHandle, hErrWrite
	Mov hErrWrite, NULL
	Or Eax, Eax
	Jz proc_error

	Invoke GlobalAlloc, LPTR, 1028
	Or Eax, Eax
	Jz proc_error
	Mov lpBuffer, Eax

pipe_read:
	Invoke ReadFile, hOutRead, lpBuffer, 1024, Addr dwRead, NULL
	Or Eax, Eax
	Jz pipe_error
	Cmp dwRead, 0
	Je pipe_error
	; Add zero to end of string
	Mov Edx, lpBuffer
	Add Edx, dwRead
	Mov Byte Ptr [Edx], 0
	; Call callback
	;Invoke CallCallback, lpCallback, 1, lpBuffer, dwRead
	DoCallback lpCallback, 1, lpBuffer, dwRead
	Jmp pipe_read

pipe_error:
	Invoke GetLastError
	Cmp Eax, ERROR_BROKEN_PIPE
	Je pipe_end
	; check other errors here...
	Jmp proc_error
pipe_end:

	; todo: wait_timeout error callback, type 3
	Invoke WaitForSingleObject, hChildProcess, INFINITE
	Or Eax, Eax
	Jnz proc_error

	Invoke GetExitCodeProcess, hChildProcess, Addr dwRet
	Or Eax, Eax
	Jz proc_error

	;Invoke CallCallback, lpCallback, 2, lpszScript, dwRet
	DoCallback lpCallback, 2, lpszScript, dwRet
	Jmp proc_exit

proc_error:
	Mov dwRet, -1
	; todo: put error code in register before jmp
	; put the value instead of the zero in the last argument
	;Invoke CallCallback, lpCallback, 3, lpszScript, 0
	DoCallback lpCallback, 3, lpszScript, 0

proc_exit:
	Cmp lpBuffer, NULL
	Je @F
	Invoke GlobalFree, lpBuffer
@@:
	Cmp hChildProcess, NULL
	Je @F
	Invoke CloseHandle, hChildProcess
@@:
	Cmp hOutReadTmp, NULL
	Je @F
	Invoke CloseHandle, hOutReadTmp
@@:
	Cmp hOutRead, NULL
	Je @F
	Invoke CloseHandle, hOutRead
@@:
	Cmp hOutWrite, NULL
	Je @F
	Invoke CloseHandle, hOutWrite
@@:
	Cmp hErrWrite, NULL
	Je @F
	Invoke CloseHandle, hErrWrite
@@:

	Mov Eax, dwRet
	Ret
CreateProcessAndRedirectOutput EndP

LaunchRedirectedProcess Proc Private lpszCmdLine:DWord, hChildStdOut:DWord, hChildStdIn:DWord, hChildStdErr:DWord, lpdwChildHandle:DWord
	Local pi:PROCESS_INFORMATION
	Local psi:STARTUPINFO
	; Set up the start up info Struct.
	Invoke RtlZeroMemory, Addr psi, SizeOf STARTUPINFO
	Mov psi.cb, SizeOf STARTUPINFO
	Mov psi.dwFlags, STARTF_USESTDHANDLES Or STARTF_USESHOWWINDOW
	m2m psi.hStdOutput, hChildStdOut
	m2m psi.hStdInput, hChildStdIn
	m2m psi.hStdError, hChildStdErr
	Mov psi.wShowWindow, SW_HIDE
	; Launch the process that you want to redirect. 
	Invoke CreateProcess, NULL, lpszCmdLine, NULL, NULL, TRUE, CREATE_NEW_CONSOLE, NULL, NULL, Addr psi, Addr pi
	Or Eax, Eax
	Jz proc_error
	; Close any unnecessary handles.
	Invoke CloseHandle, pi.hThread
	Or Eax, Eax
	Jz proc_error
	; Store the process handle in dwChildHandle
	Mov Ecx, lpdwChildHandle
	m2m DWord Ptr [Ecx], pi.hProcess
	; return non zero, last check eax was not zero and eax is unchanged from there
	;Mov Eax, 1
	Ret
proc_error:
	Xor Eax, Eax
	Ret
LaunchRedirectedProcess EndP
